import dataclasses
from collections import defaultdict
from dataclasses import dataclass
from typing import List, Dict, DefaultDict

from hatsudenki.packages.command.master.column import MasterColumn, ColumnEnum, ColumnRelation
from hatsudenki.packages.command.master.enum.loader import EnumLoader
from hatsudenki.packages.command.master.loader import MasterTableLoader
from hatsudenki.packages.command.master.table import MasterTable
from hatsudenki.packages.command.network.request import aio_http_get_json, aio_http_post_json, aio_http_put_json
from hatsudenki.packages.command.renderer.base import RenderUnit, FileRenderer
from hatsudenki.packages.command.stdout.output import IndentString, ToolOutput


@dataclass()
class GitWikiUploader:
    base_url: str
    token_str: str

    def _make_header(self):
        return {
            'PRIVATE-TOKEN': self.token_str
        }

    async def get_page_list(self):
        resp = await aio_http_get_json(self.base_url, self._make_header())
        return resp

    async def create_page(self, content: str, page_name: str):
        """
                gitlabのwikiの記事作成

                :param content: ページ内容
                :param page_name: 作成ページ名
                :return: レスポンス
                """
        body = {
            'content': content,
            'title': page_name,
            'format': 'markdown'
        }
        resp = await aio_http_post_json(f'{self.base_url}', body, self._make_header())
        return resp

    async def update_page(self, content: str, page_name: str):
        """
        gitlabのwikiの記事更新

        :param content: ページ内容
        :param page_name: 更新ページ名
        :return: レスポンス
        """
        body = {
            'title': page_name,
            'format': 'markdown',
            'content': content
        }
        slug = page_name.replace('/', '%2F')
        resp = await aio_http_put_json(f'{self.base_url}/{slug}', body, self._make_header())
        return resp


@dataclass()
class MarkdownTableRenderer:
    header: List[str]
    rows: List[Dict[str, str]] = dataclasses.field(default_factory=list)

    def append(self, row: Dict[str, str]):
        self.rows.append(row)

    def render(self):
        u = IndentString()
        u.add('|' + '|'.join(self.header) + '|')
        u.add('|' + '|'.join(['---------'] * len(self.header)) + '|')

        for row in self.rows:
            s = []
            for h in self.header:
                s.append(self._get(row.get(h, '----')))
            u.add('|' + '|'.join(s) + '|')
        return u

    def _get(self, d):
        return d if d else '----'


class MasterDataWikiRenderUnit(RenderUnit[MasterTable]):
    def render(self):
        ToolOutput.anchor(f'{self.data.label} をMarkdownに変換')
        u = IndentString()
        u.add(f'# {self.data.excel_sheet_name}')
        u.add(f'## 情報')
        u.add(self.data.description if self.data.description else '*!! descriptionが設定されていません !!*')
        u.blank_line()

        u.add(self._gen_info().render())

        u.add(f'## 担当者')
        u.add(self._gen_assigner().render())

        u.add(f'## スキーマ情報')
        u.add(self._column().render())

        ToolOutput.pop('OK')
        return u

    @property
    def page_name(self):
        return f'MasterData/{self.data.excel_name}'

    def get_page_name(self, data: MasterTable):
        return f'MasterData/{data.excel_name}'

    def _label(self, data: MasterTable):
        return f'【{data.excel_name}】{self.data.excel_sheet_name}'

    def _gen_assigner(self):
        tr = MarkdownTableRenderer(['セクション', '担当者'])
        tr.append({'セクション': 'クライアント', '担当者': ', '.join(self.data.assignee_clients)})
        tr.append({'セクション': 'サーバー', '担当者': ', '.join(self.data.assignee_servers)})
        tr.append({'セクション': 'プランナー', '担当者': ', '.join(self.data.assignee_planners)})
        return tr

    def _gen_info(self):
        tr = MarkdownTableRenderer(['情報', '値'])
        tr.append({'情報': 'テーブル名', '値': self.data.table_name})
        tr.append({'情報': 'スキーマ定義ファイル', '値': self.data.rel_path})
        tr.append({'情報': 'PACK出力', '値': self._bool_to_str(self.data.is_out_pack, '○', '☓')})
        return tr

    def _bool_to_str(self, val: bool, true_str: str, false_str: str):
        return true_str if val else false_str

    def _column(self):
        tr = MarkdownTableRenderer(['名称', 'カラム名', 'index', '型', '入力制限', 'コメント', 'PACK出力'])

        for k, c in self.data.columns.items():
            tr.append({
                '名称': c.excel_raw_header_name,
                'カラム名': c.column_name,
                'index': 'hash' if c.is_hash_key else 'range' if c.is_range_key else None,
                '型': self._type_to_str(c),
                '入力制限': self._limit_to_str(c),
                'コメント': c.comment,
                'PACK出力': self._bool_to_str(c.is_no_pack, '☓', '○')
            })
        return tr

    def _type_to_str(self, column: MasterColumn):
        if type(column) is ColumnEnum:
            return f'enum<br>{column.to}'
        if type(column) is ColumnRelation:
            return f'master<br>{column.to}'

        return column.type_name

    def _limit_to_str(self, column: MasterColumn):
        ret = []
        l = column.value_limit
        if l is None:
            return None
        for t, v in l.items():
            if t == 'choice':
                for vv in v:
                    ret.append(f'- {vv.excel_name}')
            elif t == 'master':
                ret.append(f'[[{self.get_page_name(v)}#{v.excel_sheet_name}]]')
        return '<br />'.join(ret)

    def render_index(self):
        return {
            'Book': f'[{self.data.excel_name}#{self.data.excel_sheet_name}](/{self.page_name}#{self.data.excel_sheet_name})',
            'Table': self.data.table_name,
            'Yaml': self.data.rel_path,
            'description': self.data.description,
            'Client': '<br />'.join(self.data.assignee_clients),
            'Server': '<br />'.join(self.data.assignee_servers),
            'Planner': '<br />'.join(self.data.assignee_planners)
        }


@dataclass
class MasterDataWikiRenderer(FileRenderer):
    master_loader: MasterTableLoader
    enum_loader: EnumLoader

    def render_header(self) -> IndentString:
        return super().render_header()

    def render_units(self):
        ToolOutput.anchor('Wiki出力を開始')
        u = IndentString()
        for k, table in self.master_loader.iter():
            un = MasterDataWikiRenderUnit(table).render()
            u.add(un)
        ToolOutput.pop('OK')

        return u

    def render_index(self, index: MarkdownTableRenderer):
        ToolOutput.anchor('目次')
        u = IndentString()
        u.add('# MasterData目次')
        u.add(index.render())
        return u

    async def render_to_wiki(self, base_url, token_str):
        ToolOutput.anchor('Wiki出力を開始')

        wiki = GitWikiUploader(base_url, token_str)

        page_list = await wiki.get_page_list()
        page_slug_list = [v['slug'] for v in page_list]

        d: DefaultDict[str, List[MasterDataWikiRenderUnit]] = defaultdict(list)
        ToolOutput.anchor('YAML読み込み')
        for k, table in self.master_loader.iter():
            ru = MasterDataWikiRenderUnit(table)
            ToolOutput.out(f'{ru.page_name} # {ru.data.excel_sheet_name}')
            d[ru.page_name].append(ru)
        ToolOutput.pop('OK')

        ToolOutput.anchor('書き出し')
        num = 1
        index = MarkdownTableRenderer(['ID', 'Book', 'Client', 'Server', 'Planner', 'Table'])
        for p, rus in d.items():
            ToolOutput.anchor(p)
            u = IndentString()
            for r in rus:
                u.add(r.render())
                idx = r.render_index()
                idx['ID'] = str(num)
                num += 1
                index.append(idx)

            if p in page_slug_list:
                ToolOutput.print(f'{p}更新します')
                await wiki.update_page(content=u.render(), page_name=p)
            else:
                ToolOutput.print(f'{p}作成します')
                await wiki.create_page(content=u.render(), page_name=p)
            ToolOutput.pop('OK')

        ix = self.render_index(index)

        index_title = 'MasterData/☆INDEX☆'
        if index_title in page_slug_list:
            ToolOutput.print(f'{index_title}更新します')
            await wiki.update_page(content=ix.render(), page_name=index_title)
        else:
            ToolOutput.print(f'{index_title}作成します')
            await wiki.create_page(content=ix.render(), page_name=index_title)

        ToolOutput.pop('OK')



import os
from itertools import chain
from pathlib import Path


class BaseInfo(object):
    def __init__(self, base_path: Path, full_path: Path, *args, **kwargs):
        bp = os.path.normpath(str(base_path))
        fp = os.path.normpath(str(full_path))
        self.full_path = fp
        self.rel_path: str = fp.replace(bp, '')
        self.dirs = self.rel_path.replace('.yml', '').lstrip(os.sep).split(os.sep)

        self.base_class_name = ''.join(
            [s.capitalize() for s in self.rel_path.replace('_', os.sep).split(os.sep)[2:]]).replace('.yml', '')

    @property
    def top_level(self):
        return self.dirs[0]

    @property
    def filename(self) -> str:
        return self.dirs[-1]

    @property
    def filename_without_ext(self):
        r = self.dirs[-1].split('.')
        if len(r) > 1:
            return ''.join(r[:-1])
        else:
            return r[0]

    @property
    def dirname(self) -> str:
        return os.path.dirname(self.full_path)

    @property
    def normalize_name(self):
        if len(self.dirs) > 1 and self.dirs[-1] == self.dirs[-2]:
            return '_'.join(self.dirs[0:-1])
        return '_'.join(self.dirs)

    @property
    def normalize_dirs(self):
        if len(self.dirs) > 1 and self.dirs[-1] == self.dirs[-2]:
            return self.dirs[0:-1]
        return self.dirs

    @property
    def full_classname(self):
        s = chain.from_iterable([[c.capitalize() for c in a.split('_')] for a in self.normalize_dirs])
        return ''.join(s)

    def __str__(self):
        return self.rel_path

    def __repr__(self):
        return self.rel_path



import datetime
from typing import Optional, Dict

from hatsudenki.packages.command.stdout.output import ToolOutput, snake_to_camel, IndentString


def master_column(type_str: str, python_type_str: str, cs_type_str: str, bq_type_str: str):
    """
    Masterのカラムであることを表す
    :param type_str: yaml上で設定するべきtype文字列
    :param python_type_str: python上で扱われるtype文字列
    :param cs_type_str: cs上で扱われるtype文字列
    :param bq_type_str: BigQuery上で扱われるtype文字列
    :return:
    """

    def deco(func):
        func.type_name = type_str
        func.python_type_name = python_type_str
        func._cs_type_name = cs_type_str
        func._bq_type_name = bq_type_str

        ColumnFactory.register(type_str, func)
        return func

    return deco


class ColumnFactory(object):
    """
    typeの値を元に対応したColumnインスタンスを生成する
    """
    _type_table = {}

    @classmethod
    def create(cls, table_name: str, key_name: str, data: dict, master_loader):
        try:
            type = data['type']
            return cls._type_table[type](table_name, key_name, data, master_loader)
        except Exception as e:
            ToolOutput.out_error(f'columnの生成に失敗 table={table_name}, key={key_name}')
            raise e

    @classmethod
    def register(cls, type_name, column_cls):
        cls._type_table[type_name] = column_cls


class MasterColumn(object):
    """
    Masterカラムのベース
    """

    type_name: str = ''
    python_type_name: str = ''
    _cs_type_name: str = ''
    _bq_type_name: str = ''

    def __init__(self, table_name: str, key_name: str, data: dict, master_loader):
        from hatsudenki.packages.command.master.enum.loader import EnumLoader
        from hatsudenki.packages.command.master.loader import MasterTableLoader
        self.table_name = table_name
        self.column_name = key_name
        self.data = data
        self.master_loader: MasterTableLoader = master_loader
        self.enum_loader: EnumLoader = self.master_loader.enum_loader

    @property
    def cs_type_name(self):
        return self.__class__._cs_type_name

    @property
    def bq_type_name(self):
        return self.__class__._bq_type_name

    @property
    def is_relation(self):
        return False

    @property
    def excel_header_name(self):
        return self.data.get('name', self.column_name)

    @property
    def excel_raw_header_name(self):
        return self.data.get('name', self.column_name)

    @property
    def table(self):
        return self.master_loader.get_by_table_name(self.table_name)

    @property
    def python_name(self):
        return self.column_name

    @property
    def python_define_name(self):
        return f'{self.python_name} = column.{self.python_type_name}(name="{self.python_name}")'

    @property
    def python_init_name(self):
        return f"self.{self.python_name} = fields.{self.python_name}.convert(kwargs.get(fields.{self.python_name}.name))"

    @property
    def cs_name(self):
        """
        CS側の変数名

        :return: str
        """
        return snake_to_camel(self.python_name)

    @property
    def cs_argument_name(self):
        """
        CS側の引数名

        :return: str
        """
        return f'{self.cs_type_name} {self.column_name}'

    @property
    def cs_assign_name(self):
        return f'{self.cs_name} = {self.column_name};'

    @property
    def cs_define_name(self):
        return f'public {self.cs_type_name} {self.cs_name} {{ get; set; }}'

    @property
    def is_hash_key(self):
        return self.data.get('hash', False)

    @property
    def is_range_key(self):
        return self.data.get('range', False)

    @property
    def is_serial_key(self):
        return self.data.get('is_serial', False)

    @property
    def is_no_pack(self):
        if not self.table.is_out_pack:
            # テーブル自体がPackをスキップする設定の場合は無条件にスキップ
            return True
        return self.data.get('no_pack', False)

    @property
    def comment(self):
        return self.data.get('comment', None)

    def generate_value(self, value: any):
        return value

    def reverse_value(self, value: any):
        return value

    def cs_access_name(self):
        return None

    @property
    def value_limit(self) -> Optional[Dict[str, any]]:
        return None

    @property
    def serial_resolver_name(self):
        if not self.is_serial_key:
            return None

        serial = f"'{self.table_name}_'" + f' + str(self.{self.column_name}).zfill(10)'

        i = IndentString()

        i.add('@property')
        i.indent('def log_id(self):')
        i.add(f'return {serial}')

        return i

    @staticmethod
    def serial_resolver_default():
        i = IndentString()

        i.add('@property')
        i.indent('def log_id(self):')
        i.add('return self.one_cursor')

        return i


@master_column('number', 'MasterColumnInt', 'uint', 'INTEGER')
class ColumnNumber(MasterColumn):
    def generate_value(self, value: any):
        info: str = f'({self.table.excel_name}.xlsx, sheet: {self.table.excel_sheet_name}, column: {self.excel_header_name})'
        if type(value) == str:
            try:
                return int(value)
            except ValueError as e:
                raise Exception(f'数値カラムに文字列が入っている({info})')

        if value < 0:
            raise Exception(f'uint数値カラムに負数が入っている({info})')
        if type(value) != int:
            raise Exception(f'数値カラムに整数ではない値が入っている。小数？({info})')
        return value


@master_column('string', 'MasterColumnString', 'string', 'STRING')
class ColumnString(MasterColumn):
    def generate_value(self, value: any):
        if value is not None:
            return str(value)

        return ''


@master_column('master', 'MasterColumnRelation', 'string', 'STRING')
class ColumnRelation(MasterColumn):

    @property
    def cs_type_name(self):
        target = self.get_target_table()
        if target.range_key is not None:
            return 'string'
        else:
            return target.hash_key.cs_type_name

    @property
    def bq_type_name(self):
        target = self.get_target_table()
        if target.range_key is not None:
            return 'string'
        else:
            return target.hash_key.bq_type_name

    @property
    def is_relation(self):
        return True

    @property
    def excel_header_name(self):
        n = self.data.get('name', self.column_name)
        to = self.to
        self_table = self.table
        target_table = self.master_loader.get_by_table_name(to)
        if self_table.excel_name == target_table.excel_name:
            return f'=HYPERLINK("#{target_table.excel_sheet_name}!A1", "{n}")'
        else:
            return f'=HYPERLINK("{target_table.excel_name}.xlsx#{target_table.excel_sheet_name}!A1", "{n}")'

    @property
    def to(self):
        return self.data['to']

    def get_target_table(self):
        return self.master_loader.get_by_table_name(self.to)

    @property
    def python_define_name(self):
        return f'{self.python_name} = column.{self.python_type_name}(name="{self.python_name}", to="{self.data["to"]}")'

    @property
    def resolver_name(self):
        u = IndentString()
        u.add('@property')
        u.indent(f'def resolved_{self.python_name}(self):')

        t = self.get_target_table()
        if t.range_key is None:
            u.add(f'a = {self.get_target_table().class_name}.get(self.{self.python_name})')
            u.indent('if a is None:')
            u.add('return None')
            u.outdent()
            u.indent('if a.is_nothing:')
            u.add('return None')
            u.outdent()
            u.add('return a')
        else:
            u.add(f'a = {self.get_target_table().class_name}.find(self.{self.python_name})')
            u.indent('if len(a) > 0 and a[0].is_nothing:')
            u.add('return None')
            u.outdent()
            u.add('return a')
        return u

    def cs_access_name(self):
        u = IndentString()
        t = self.get_target_table()

        u.add(f'public {t.class_name} {self.cs_name}Table()')
        u.indent('{')
        u.add(f'return {t.class_name}.Instance;')
        u.outdent('}')
        return u

    @property
    def value_limit(self) -> Optional[Dict[str, any]]:
        t = self.get_target_table()
        return {
            'master': t
        }


@master_column('enum', 'MasterColumnEnum', 'uint', 'INTEGER')
class ColumnEnum(MasterColumn):
    @property
    def is_relation(self):
        return True

    @property
    def excel_header_name(self):
        n = self.data.get('name', self.column_name)
        to = self.data['to']
        target_table = self.enum_loader.get_by_name(to)
        return f'=HYPERLINK("pg_excel/定義タイプ.xlsx#{target_table.excel_sheet_name}!A1", "{n}")'

    @property
    def to(self):
        return self.data['to']

    def resolve_relation(self):
        return self.enum_loader.get_by_name(self.to)

    def reverse_value(self, value: any):
        en = self.resolve_relation()
        v = en.find(value, is_raw=True)
        if v is None:
            l = f'{self.table.excel_name}.xlsx {self.excel_raw_header_name}'
            raise Exception(f'{self.table_name}: {en.excel_sheet_name} から {value} が見つかりません\n{l}')
        return v.excel_name

    def generate_value(self, value):
        en = self.resolve_relation()
        excel = en.find(value, is_raw=False)
        if excel is None:
            l = f'{self.table.excel_name}.xlsx {self.table.excel_sheet_name} {self.excel_raw_header_name}'
            raise Exception(f'enum not found {self.to} {value} {l}')
        return excel.id

    @property
    def value_limit(self):
        en = self.resolve_relation()
        return {
            'choice': en.values
        }


@master_column('date', 'MasterColumnDate', 'uint', 'TIMESTAMP')
class ColumnDatetime(MasterColumn):
    def generate_value(self, value: any):
        if type(value) is datetime.datetime:
            return int(value.timestamp())
        if value == None:
            return 0
        t = f'{self.table.excel_name}.xlsx {self.table.excel_sheet_name} {self.excel_raw_header_name}'
        raise Exception(f'invalid datetime {value} {t}')

    def reverse_value(self, value: any):
        return datetime.datetime.fromtimestamp(value)


@master_column('select', 'MasterColumnSelect', 'uint', 'INTEGER')
class ColumnSelect(ColumnEnum):
    @property
    def label(self):
        return f'[Select] {self.table.label}:{self.excel_raw_header_name}({self.column_name})'

    def resolve_target_name(self, val, is_raw=False) -> str:
        enum_data = self.resolve_relation()
        v = enum_data.find(val, is_raw=is_raw)
        if v is None:
            raise Exception(f'{self.label} から {val} が見つかりません is_raw={is_raw}')
        return v.data['value']

    def resolve_target(self, val, is_raw=False):
        n = self.resolve_target_name(val, is_raw)
        if n.startswith('enum_'):
            # enum_から始まっているターゲットはenumとして扱う
            return self.enum_loader.get_by_name(n)
        else:
            return self.master_loader.get_by_table_name(n)


@master_column('chose', 'MasterColumnChose', 'string', 'STRING')
class ColumnChose(MasterColumn):
    @property
    def selector(self) -> str:
        return f'{self.data["selector"]}'

    @property
    def label(self):
        return f'[Chose] {self.table.label}:{self.excel_raw_header_name}({self.column_name})'

    @property
    def selector_column(self) -> ColumnSelect:
        sc = self.table.columns.get(self.selector, None)
        if sc is None:
            raise Exception(f'ChoseColumnの解決に失敗 {self.table.label} から {self.selector} カラム が見つかりません')
        if isinstance(sc, ColumnSelect):
            return sc
        else:
            raise Exception(f'ChoseColumnの解決に失敗 {self.table.label} の {self.selector} カラム は Selectorではありません')

    def _resolve_value(self, s, val, is_raw=False, ret_ref_name=False):
        from hatsudenki.packages.command.master.enum.data import EnumData
        from hatsudenki.packages.command.master.table import MasterTable
        if isinstance(s, EnumData):
            en_data = s.find(val, is_raw)
            if en_data is None:
                raise Exception(f'{self.label} の解決に失敗 {s.label} から {val} が見つかりません')
            if ret_ref_name:
                return en_data.excel_name
            else:
                return en_data.id
        elif isinstance(s, MasterTable):
            return val
        else:
            raise Exception(f'{self.label} selectorの戻り値が不正です')

    def resolver_name(self, selector: ColumnSelect):
        u = IndentString()
        te = selector.resolve_relation()

        u.add('@property')
        u.indent(f'def resolved_{self.python_name}(self):')
        u.add('from common.enums import def_enum')
        u.add(f'tbl = def_enum.Enum{te.full_classname}.resolve(self.{selector.python_name})')
        u.add(f'return tbl.get(self.{self.python_name})')
        return u

    def resolve_target(self, val, is_raw=False):
        sc = self.selector_column
        return sc.resolve_target(val, is_raw)

    def resolve_value(self, sel, value, is_raw=False, ret_ref_name=False):
        t = self.resolve_target(sel, is_raw)
        return self._resolve_value(t, value, is_raw, ret_ref_name)

    def generate_value(self, value: any):
        return value

    def reverse_value(self, value: any):
        raise Exception('ChoseColumnはSelectorの値が無いと確定しません')



from typing import Type, Union

from hatsudenki.packages.expression.condition import KeyConditionExpression, FilterConditionExpression
from hatsudenki.packages.table.multi import MultiHatsudenkiTable
from hatsudenki.packages.table.solo import SoloHatsudenkiTable


class DirectFind(object):
    def __init__(self, model: Union[Type[SoloHatsudenkiTable], Type[MultiHatsudenkiTable]]):
        self.model = model
        self.cond = KeyConditionExpression()
        self.filter = FilterConditionExpression()
        self.limit = 0

    def set_key(self, hash_val, range_val):
        with self.cond:
            self.cond.equal(self.model.get_hash_key_name(), hash_val)
            if range_val is not None:
                self.cond.op_and()
                self.cond.equal(self.model.get_range_key_name(), range_val)

    def set_key_direct(self, **vals):
        f = False
        with self.cond:
            for k, v in vals.items():
                if not f:
                    self.cond.op_and()
                fc = self.model.get_field_class(k)
                if fc is None:
                    raise Exception(f'invalid key {k}')
                self.cond.equal(k, fc.serialize(v), raw=True)

    def set_limit(self, limit: int):
        self.limit = limit

    async def exec(self):
        pass
        # res = await self.model.query(self.cond, self.filter, limit=self.limit)
        # return res



from logging import getLogger
from typing import Dict, Union

from hatsudenki.packages.cache.inmemory.table import InMemoryTableSolo, InMemoryTableMulti, SyncInMemoryTableSolo, \
    SyncInMemoryTableMulti

_logger = getLogger(__name__)


class InMemoryTableManager(object):
    _all_tables: Dict[
        str, Union[InMemoryTableSolo, InMemoryTableMulti, SyncInMemoryTableSolo, SyncInMemoryTableMulti]] = {}

    @classmethod
    def register(cls, tbl):
        tbl_name = tbl.Meta.table_name
        if tbl_name is None:
            _logger.warning(f'table name not set. {tbl}')
            return

        if tbl_name in cls._all_tables:
            _logger.critical(f'table name duplicate! {tbl_name}')
            return
        _logger.info(f'register in-memory table {tbl_name}')
        cls._all_tables[tbl_name] = tbl

    @classmethod
    def get_by_table_name(cls, table_name: str):
        return cls._all_tables[table_name]

    @classmethod
    def iter(cls):
        return cls._all_tables.items()

    @classmethod
    def invalidate_all(cls):
        for k, t in cls.iter():
            t.invalidate()



from hatsudenki.packages.command.master.enum.data import EnumData
from hatsudenki.packages.command.renderer.base import RenderUnit, FileRenderer
from hatsudenki.packages.command.stdout.output import IndentString


class CSEnumRenderUnit(RenderUnit[EnumData]):
    def render(self):
        u = IndentString()

        if self.data.is_out_desc:
            u.add(f'using System.ComponentModel;')
            u.blank_line()

        u.add(f'public enum Enum{self.data.full_classname}')
        u.indent('{')
        for v in self.data.values:
            if self.data.is_out_desc:
                u.add(f'{v.cs_desc}')
            u.add(f'{v.cs_str}')

        u.add(f'MAX = {len(self.data.values)},')
        u.outdent('}')
        return u


class CSEnumRenderer(FileRenderer):
    pass



from logging import getLogger
from typing import Type, TypeVar, Iterator, List

from hatsudenki.packages.cache.base.memory.solo import CacheBaseTableSolo

T = TypeVar('T')

_logger = getLogger(__name__)


class SyncDynamicCacheBaseTableSolo(CacheBaseTableSolo):
    _mapped = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    def build(cls: Type[T]) -> List[T]:
        return []

    @classmethod
    def _setup(cls):
        if cls.check_invalidate():
            _logger.info(f'setup {cls.Meta.table_name}')
            cls.prepare()
            r = cls.build()
            cls._set_records(r)

    @classmethod
    def check_invalidate(cls):
        return cls._mapped is None

    @classmethod
    def invalidate(cls):
        cls._mapped = None

    @classmethod
    def prepare(cls):
        cls._mapped = {}

    @classmethod
    def get(cls: Type[T], hash_key, range_key=None) -> T:
        """
        HASHキーを指定して一件取得

        :param hash_key: ハッシュキー値
        :param range_key: レンジキー値（Soloテーブルはこの値は必ずNoneになる）
        :return: 条件にマッチしたレコード
        """

        cls._setup()
        return cls._get(hash_key, range_key)

    @classmethod
    def get_by_cursor(cls: Type[T], cursor: str) -> T:
        """
        カーソル文字列から一件取得

        :param cursor: カーソル文字列
        :return: 条件にマッチしたレコード
        """
        cls._setup()
        return cls._get_by_cursor(cursor)

    @classmethod
    def check_exist_hash(cls, hash_key):
        """
        指定されたハッシュキーのレコードがあるか判定する

        :param hash_key: 検索対象値
        :return: bool
        """
        cls._setup()
        return cls._check_exist_hash(hash_key)

    @classmethod
    def iter(cls: Type[T]) -> Iterator[T]:
        """
        全件列挙イテレータ

        :return: iter
        """

        cls._setup()
        return cls._iter()



from pathlib import Path
from typing import Dict

from openpyxl.styles import Font, Color
from openpyxl.utils import coordinate_to_tuple

from hatsudenki.packages.command.excel.excel_data import ExcelData, CellStyle
from hatsudenki.packages.command.master.table import MasterTable
from hatsudenki.packages.command.stdout.output import ToolOutput


class MasterExcelBook(ExcelData):
    TABLE_NAME_SHEET_NAME = 'テーブル名'
    RELATION_SHEET_NAME = 'リレーション'

    def __init__(self, file_path: Path, book_info: Dict[str, MasterTable]):
        super().__init__(file_path)
        self.sort_dict = {}
        self.table_infos = book_info
        if file_path.exists():
            self.load()
        else:
            self.create()

        s = self.book.named_styles
        if CellStyle.RED.name not in s:
            self.book.add_named_style(CellStyle.RED)
        if CellStyle.BLUE.name not in s:
            self.book.add_named_style(CellStyle.BLUE)
        if CellStyle.GREEN.name not in s:
            self.book.add_named_style(CellStyle.GREEN)

        self._generate_table_sheets()
        self.book._sheets.sort(
            key=lambda t: self.sort_dict[t.title] if t.title in self.sort_dict else 0,
            reverse=True
        )

    def _generate_table_sheets(self):
        """
        マスターテーブルに対応するシートを生成
        """
        origin = 0
        self.sort_dict = {}
        # 所属するテーブルでループ
        for idx, (sheet_name, table) in enumerate(self.table_infos.items()):
            ToolOutput.anchor(f'[{sheet_name}]シート作成開始...')

            if sheet_name in self.book.sheetnames:
                # すでに対応するシートが存在する
                sheet = self.book[sheet_name]
                old_header = self.get_header_values(sheet_name)
                self.sort_dict[sheet_name] = table.data['priority'] if 'priority' in table.data else 0
            else:
                sheet = self.book.create_sheet(sheet_name, origin + idx)
                ToolOutput.out(f'[{sheet_name}]シート を追加します')
                old_header = {}
                self.sort_dict[sheet_name] = table.data['priority'] if 'priority' in table.data else 0
            tail_coord = old_header.get('TAG', 'A1')

            # カラム列の生成
            for idx_2, (k, c) in enumerate(table.columns.items()):
                # すでに列が存在するか？
                coord = old_header.get(c.excel_raw_header_name, None)
                if coord is None:
                    # 新規に追加された列である
                    ToolOutput.out(f'{c.excel_raw_header_name} 列が[{tail_coord}]に追加されました')

                    row_idx, col_idx = coordinate_to_tuple(tail_coord)
                    sheet.insert_cols(col_idx, 1)
                    cell = sheet[tail_coord]
                    tail_coord = cell.offset(row=0, column=1).coordinate
                else:
                    cell = sheet[coord]

                cell.style = CellStyle.RED.name
                # 特殊なカラムの処理
                if c.is_no_pack:
                    # packに出力しない列
                    cell.style = CellStyle.GREEN.name
                if c.is_relation:
                    # 参照が設定されている列
                    cell.value = c.excel_header_name
                    cell.font = Font(name="Meiryo UI", size=10, underline="single",
                                     color=Color(rgb=None, indexed=None, auto=None, theme=10, tint=0.0, type="theme"))
                else:
                    # 普通の列
                    cell.value = c.excel_header_name
            ToolOutput.pop('OK')

            cell = sheet[tail_coord]
            cell.style = CellStyle.BLUE.name
            cell.value = 'TAG'



from typing import List, Type, TypeVar

from hatsudenki.define.config import CURSOR_SEPARATOR
from hatsudenki.packages.expression.condition import ConditionExpression
from hatsudenki.packages.table.define import TableType
from hatsudenki.packages.table.solo import SoloHatsudenkiTable

T = TypeVar('T')


class MultiHatsudenkiTable(SoloHatsudenkiTable):

    @classmethod
    def get_table_type(cls):
        if cls.Meta.is_root:
            return TableType.RootTable
        return TableType.SingleMultiTable

    @classmethod
    def _is_allow_modify(cls, key: str):
        t = super()._is_allow_modify(key)
        if t:
            return cls.get_range_key_name() != key
        return False

    @property
    def range_key_name(self):
        return self.__class__.get_range_key_name()

    @classmethod
    def get_range_key_name(cls):
        p = cls.get_primary_index()
        return p.range_key

    @property
    def range_value(self):
        """
        ハッシュキーの値を取得

        :return: ハッシュキーの値（型はモデルに依存）
        """
        return getattr(self, self.range_key_name)

    @property
    def one_cursor(self):
        r = self.__class__.get_field_class(self.range_key_name)
        k = super().one_cursor + CURSOR_SEPARATOR + r.to_string(self.range_value)
        return k

    @property
    def serialized_key(self):
        c = self.__class__
        hk = c._hash_key_name
        rk = c._range_key_name
        d = self.__dict__

        return {
            hk: c._serializer[hk](d[hk]),
            rk: c._serializer[rk](d[rk])
        }

    @classmethod
    def not_exist_condition(cls, cond: ConditionExpression = None):
        p = super().not_exist_condition(cond)
        p.op_and()
        p.attribute_not_exists(cls.get_range_key_name())
        return p

    @classmethod
    def exist_condition(cls, cond: ConditionExpression = None):
        p = super().exist_condition(cond)
        p.op_and()
        p.attribute_exists(cls.get_range_key_name())
        return p

    async def reget(self):
        c = await self.__class__.get(self.hash_value, self.range_value)
        self.copy(c)

    @classmethod
    def get_serialized_key(cls, hash_val: any, range_val: any = None):
        if range_val is None:
            raise Exception('invalid key')
        hash_key = cls._hash_key_name
        range_key = cls._range_key_name
        return {
            hash_key: cls._serializer[hash_key](hash_val),
            range_key: cls._serializer[range_key](range_val)
        }

    @property
    def many_cursor(self):
        return f'{self.hash_value}//{self.range_value}'

    def set_range_key(self, val):
        """
        ハッシュキーの値をセットする
        このメソッドを使わずにハッシュキーを書き換えると例外が発生する

        :param val: セットする値（型はモデルに依存）
        :return: None
        """
        rk = self.get_range_key_name()
        f = self.__class__.get_field_class(rk)
        self.force_set_key(rk, f.get_data(val))

    @classmethod
    async def query_by_cursor(cls: Type[T], cursor: str) -> T:

        rp = cursor.split(CURSOR_SEPARATOR)

        # k = cls.get_serialized_key(*rp[:2])
        return await cls.get(*rp[:2])

    @classmethod
    async def query_list_by_cursor(cls: Type[T], cursor: str) -> List[T]:
        return await cls.query_list({
            cls.get_hash_key_name(): cursor
        })

    @classmethod
    def get_range_field_class(cls):
        return cls.get_field_class(cls.get_range_key_name())

    @classmethod
    def get_primary_key_names(cls):
        return [cls.get_hash_key_name(), cls.get_range_key_name()]

    async def remove(self):
        await self.__class__.delete(self.hash_value, self.range_value)

    @classmethod
    def _check_key(cls, hash_val: any, range_val: any):
        if range_val is None:
            raise Exception('invalid key')



from hatsudenki.define.config import TAG_SEPARATOR
from hatsudenki.packages.command.hatsudenki.field.util import field
from hatsudenki.packages.command.hatsudenki.field.base import HatsudenkiFieldBase
from hatsudenki.packages.command.stdout.output import snake_to_camel, IndentString


@field('uuid', 'uuid.UUID')
class UUIDField(HatsudenkiFieldBase):
    PythonStr = 'field.UUIDField'

    @property
    def resolver_setter_name(self):
        return '{val.hex}'


@field('date', 'datetime')
class DateField(HatsudenkiFieldBase):
    PythonStr = 'field.DateField'

    def _parse_opt(self):
        o = super()._parse_opt()
        t = self.data.get('ttl', None)
        if t is not None and t is True:
            o.append(f'ttl=True')
        return o


@field('update_date', 'datetime')
class UpdateDateField(HatsudenkiFieldBase):
    PythonStr = 'field.UpdateDateField'


@field('create_date', 'datetime')
class CreateDateField(HatsudenkiFieldBase):
    PythonStr = 'field.CreateDateField'


@field('enum', 'int')
class EnumField(HatsudenkiFieldBase):
    PythonStr = 'field.EnumField'

    def gen_init_str(self, direct_assign=True):
        return super().gen_init_str(direct_assign)

    def _parse_opt(self):
        o = super()._parse_opt()
        o.append(f"to=def_enum.{self.target_enum_name}")
        return o

    @property
    def target_enum_name(self):
        return f"Enum{snake_to_camel(self.data['to'])}"

    @property
    def resolver_string(self):
        if not self.is_alias_key:
            return None
        if self.is_range:
            # 子テーブル且つレンジキーの場合はエイリアスキーとなる
            l = len(self.parent.tag_name) + len(TAG_SEPARATOR)
            parent = self.parent.parent_table
            parent_rk_name = parent.range_key.name

            body = IndentString()

            # getter
            body.add('@property')
            body.indent(f'def {self.name}(self) -> def_enum.{self.target_enum_name}:')
            body.add(f'return self.Meta.key_alias_type.to(int(self.{parent_rk_name}[{l}:]))')
            body.outdent()
            body.blank_line()
            # setter
            body.add(f'@{self.name}.setter')
            body.indent(f'def {self.name}(self, val):')
            body.indent('if val is None:')
            body.add('return')
            body.outdent()
            body.add(f'self.{parent_rk_name} = f"{self.parent.tag_name}{TAG_SEPARATOR}{{val.value}}"')
            body.outdent()
            body.outdent()

            return body
        return None



from pathlib import Path
from typing import Dict

from hatsudenki.packages.command.files import yaml_load
from hatsudenki.packages.command.master.column import MasterColumn, ColumnFactory
from hatsudenki.packages.command.stdout.output import snake_to_camel
from hatsudenki.packages.command.util.base_info import BaseInfo


class MasterTable(BaseInfo):
    """
    Masterデータのスキーマ定義
    """

    def __init__(self, base_path: Path, full_path: Path, table_loader, *args, **kwargs):
        from hatsudenki.packages.command.master.loader import MasterTableLoader
        super().__init__(base_path, full_path, *args, **kwargs)
        self.data = yaml_load(full_path)
        self.table_loader: MasterTableLoader = table_loader
        self.columns: Dict[str, MasterColumn] = {}
        self.shadow_columns: Dict[str, MasterColumn] = {}
        self._build_columns()
        self.hash_key = next((c for c in self.columns.values() if c.is_hash_key), None)
        if self.hash_key is None:
            raise Exception(f'{self.table_name} に hashキーが設定されていません "{full_path}:1"')
        self.range_key = next((c for c in self.columns.values() if c.is_range_key), None)

        self.serial_key = next((c for c in self.columns.values() if c.is_serial_key), None)

    def _build_columns(self):
        self.columns = {k: ColumnFactory.create(self.table_name, k, c, self.table_loader) for k, c in
                        self.data.get('column', {}).items()}

        self.shadow_columns = {k: ColumnFactory.create(self.table_name, k, c, self.table_loader) for k, c in
                               self.data.get('shadow_column', {}).items()}

    @property
    def table_name(self):
        return 'master_' + self.normalize_name

    @property
    def class_name(self):
        return 'Master' + snake_to_camel(self.normalize_name)

    @property
    def excel_sheet_name(self):
        return self.data.get('name', self.table_name)

    @property
    def excel_name(self):
        return self.data['excel']

    @property
    def label(self):
        return f'{self.rel_path}【{self.excel_name}】{self.excel_sheet_name}'

    @property
    def reference_column(self):
        r = next((c for c in self.columns.values() if c.column_name == 'name'), None)
        if r is None:
            return self.columns[next(self.columns.keys().__iter__())]
        return r

    @property
    def cursor_labels(self):
        ret = [self.hash_key]
        if self.range_key is not None:
            ret.append(self.range_key)

        return ret

    @property
    def assignee(self):
        return self.data.setdefault('assignee', {})

    @property
    def assignee_clients(self):
        return self.assignee.get('clients', [])

    @assignee_clients.setter
    def assignee_clients(self, val):
        self.assignee['clients'] = val

    @property
    def assignee_servers(self):
        return self.assignee.get('servers', [])

    @assignee_servers.setter
    def assignee_servers(self, val):
        self.assignee['servers'] = val

    @property
    def assignee_planners(self):
        return self.assignee.get('planners', [])

    @assignee_planners.setter
    def assignee_planners(self, val):
        self.assignee['planners'] = val

    @property
    def is_out_pack(self):
        return self.data.get('out_pack', False)

    @property
    def description(self):
        return self.data.get('description', None)

    def find_column_by_excel_name(self, excel_name: str):
        for k, c in self.columns.items():
            if c.excel_raw_header_name == excel_name:
                return c
